/**********************************************************************
Copyright (c) 2008 Erik Bengtson and others. All rights reserved.
Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

Contributors :
2008 Andy Jefferson - change to use ExcelUtils
2008 Andy Jefferson - mechanism for persisting types as String
2011 Andy Jefferson - clean up, and add support for Maps
 ...
***********************************************************************/
package org.datanucleus.store.excel.fieldmanager;

import java.lang.reflect.Array;
import java.util.Calendar;
import java.util.Collection;
import java.util.Date;
import java.util.Iterator;
import java.util.Map;

import org.apache.poi.ss.usermodel.Cell;
import org.apache.poi.ss.usermodel.CreationHelper;
import org.apache.poi.ss.usermodel.Row;
import org.datanucleus.ClassLoaderResolver;
import org.datanucleus.metadata.AbstractClassMetaData;
import org.datanucleus.metadata.AbstractMemberMetaData;
import org.datanucleus.metadata.ColumnMetaData;
import org.datanucleus.metadata.IdentityType;
import org.datanucleus.metadata.Relation;
import org.datanucleus.store.ExecutionContext;
import org.datanucleus.store.ObjectProvider;
import org.datanucleus.store.excel.ExcelUtils;
import org.datanucleus.store.fieldmanager.AbstractFieldManager;
import org.datanucleus.store.types.ObjectLongConverter;
import org.datanucleus.store.types.ObjectStringConverter;
import org.datanucleus.util.Base64;
import org.datanucleus.util.NucleusLogger;

/**
 * FieldManager to handle the store information into an Excel worksheet row using an object.
 */
public class StoreFieldManager extends AbstractFieldManager
{
    protected final ObjectProvider op;

    protected final Row row;

    protected boolean insert; // Either insert or update

    public StoreFieldManager(ObjectProvider op, Row row, boolean insert)
    {
        this.op = op;
        this.row = row;
        this.insert = insert;

        // Add PK field(s) cell, so that the row is detected by ExcelUtils.getNumberOfRowsInSheetOfWorkbook
        AbstractClassMetaData cmd = op.getClassMetaData();
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            int[] pkFieldNumbers = cmd.getPKMemberPositions();
            for (int j=0;j<pkFieldNumbers.length;j++)
            {
                int colNumber = (int)ExcelUtils.getColumnIndexForFieldOfClass(cmd, pkFieldNumbers[j]);
                if (row.getCell(colNumber) == null)
                {
                    row.createCell(colNumber);
                }
            }
        }
        else if (op.getClassMetaData().getIdentityType() == IdentityType.DATASTORE)
        {
            int datastoreIdColNumber = (int)ExcelUtils.getColumnIndexForFieldOfClass(cmd, -1);
            if (row.getCell(datastoreIdColNumber) == null)
            {
                row.createCell(datastoreIdColNumber);
            }
        }
    }

    protected boolean isStorable(int fieldNumber)
    {
        AbstractMemberMetaData mmd = op.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        return isStorable(mmd);
    }

    protected boolean isStorable(AbstractMemberMetaData mmd)
    {
        if ((insert && mmd.isInsertable()) || (!insert && mmd.isUpdateable()))
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    protected int getColumnIndexForMember(int memberNumber)
    {
        return ExcelUtils.getColumnIndexForFieldOfClass(op.getClassMetaData(), memberNumber);
    }

    public void storeBooleanField(int fieldNumber, boolean value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeByteField(int fieldNumber, byte value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeCharField(int fieldNumber, char value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeDoubleField(int fieldNumber, double value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeFloatField(int fieldNumber, float value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeIntField(int fieldNumber, int value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeLongField(int fieldNumber, long value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue((double)value);
    }

    public void storeShortField(int fieldNumber, short value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        cell.setCellValue(value);
    }

    public void storeStringField(int fieldNumber, String value)
    {
        if (!isStorable(fieldNumber))
        {
            return;
        }
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        if (value == null)
        {
            row.removeCell(cell);
        }
        else
        {
            CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
            cell.setCellValue(createHelper.createRichTextString(value));
        }
    }

    public void storeObjectField(int fieldNumber, Object value)
    {
        ExecutionContext ec = op.getExecutionContext();
        ClassLoaderResolver clr = ec.getClassLoaderResolver();
        AbstractMemberMetaData mmd = 
            op.getClassMetaData().getMetaDataForManagedMemberAtAbsolutePosition(fieldNumber);
        if (!isStorable(mmd))
        {
            return;
        }

        // Special cases
        int relationType = mmd.getRelationType(clr);
        if (Relation.isRelationSingleValued(relationType) && mmd.isEmbedded())
        {
            // Persistable object embedded into this table
            Class embcls = mmd.getType();
            AbstractClassMetaData embcmd = ec.getMetaDataManager().getMetaDataForClass(embcls, clr);
            if (embcmd != null) 
            {
                ObjectProvider embSM = null;
                if (value != null)
                {
                    embSM = ec.findObjectProviderForEmbedded(value, op, mmd);
                }
                else
                {
                    embSM = ec.newObjectProviderForEmbedded(mmd, embcmd, op, fieldNumber);
                }

                embSM.provideFields(embcmd.getAllMemberPositions(), new StoreEmbeddedFieldManager(embSM, row, mmd, insert));
                return;
            }
        }

        storeObjectFieldInCell(fieldNumber, value, mmd, clr);
    }

    protected void storeObjectFieldInCell(int fieldNumber, Object value, AbstractMemberMetaData mmd, ClassLoaderResolver clr)
    {
        int relationType = mmd.getRelationType(clr);
        int index = getColumnIndexForMember(fieldNumber);
        Cell cell = row.getCell((int)index);
        if (cell == null)
        {
            cell = row.createCell((int)index);
        }
        if (value == null)
        {
            row.removeCell(cell);
        }
        else if (relationType == Relation.NONE)
        {
            if (Number.class.isAssignableFrom(mmd.getType()))
            {
                cell.setCellValue(((Number)value).doubleValue());
            }
            else if (Character.class.isAssignableFrom(mmd.getType()))
            {
                CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                cell.setCellValue(createHelper.createRichTextString("" + value));
            }
            else if (Boolean.class.isAssignableFrom(mmd.getType()))
            {
                cell.setCellValue(((Boolean)value).booleanValue());
            }
            else if (Date.class.isAssignableFrom(mmd.getType()))
            {
                cell.setCellValue((Date)value);
            }
            else if (Calendar.class.isAssignableFrom(mmd.getType()))
            {
                cell.setCellValue((Calendar)value);
            }
            else if (Enum.class.isAssignableFrom(mmd.getType()))
            {
                // TODO Allow persist as ordinal value
                CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                cell.setCellValue(createHelper.createRichTextString(((Enum)value).name()));
            }
            else if (byte[].class == mmd.getType())
            {
                String strValue = new String(Base64.encode((byte[]) value));
                cell.setCellValue(strValue);
            }
            else
            {
                // Try to persist using converters
                boolean useLong = false;
                ColumnMetaData[] colmds = mmd.getColumnMetaData();
                if (colmds != null && colmds.length == 1)
                {
                    String jdbc = colmds[0].getJdbcType();
                    if (jdbc != null && (jdbc.equalsIgnoreCase("int") || jdbc.equalsIgnoreCase("integer")))
                    {
                        useLong = true;
                    }
                }

                // See if we can persist it using object converters
                ObjectStringConverter strConv = 
                    op.getExecutionContext().getNucleusContext().getTypeManager().getStringConverter(value.getClass());
                ObjectLongConverter longConv =
                    op.getExecutionContext().getNucleusContext().getTypeManager().getLongConverter(value.getClass());
                if (useLong)
                {
                    if (longConv != null)
                    {
                        cell.setCellValue(longConv.toLong(value));
                        return;
                    }
                }
                else
                {
                    if (strConv != null)
                    {
                        CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                        cell.setCellValue(createHelper.createRichTextString(strConv.toString(value)));
                        return;
                    }
                    else if (longConv != null)
                    {
                        cell.setCellValue(longConv.toLong(value));
                        return;
                    }
                }

                NucleusLogger.PERSISTENCE.warn(
                    "DataNucleus doesnt currently support persistence of field " + mmd.getFullFieldName() + 
                    " type=" + value.getClass().getName() + " - ignoring");
            }
        }
        else if (Relation.isRelationSingleValued(relationType))
        {
            // Persistable object - persist the related object and store the identity in the cell
            Object valuePC = op.getExecutionContext().persistObjectInternal(value, op, fieldNumber, -1);
            Object valueId = op.getExecutionContext().getApiAdapter().getIdForObject(valuePC);
            CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
            cell.setCellValue(createHelper.createRichTextString("[" + valueId.toString() + "]"));
        }
        else if (Relation.isRelationMultiValued(relationType))
        {
            // Collection/Map/Array
            if (mmd.hasCollection())
            {
                StringBuffer cellValue = new StringBuffer("[");
                Collection coll = (Collection)value;
                Iterator collIter = coll.iterator();
                while (collIter.hasNext())
                {
                    Object element = collIter.next();
                    Object elementPC = op.getExecutionContext().persistObjectInternal(element, op, fieldNumber, -1);
                    Object elementID = op.getExecutionContext().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (collIter.hasNext())
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                cell.setCellValue(createHelper.createRichTextString(cellValue.toString()));
            }
            else if (mmd.hasMap())
            {
                AbstractClassMetaData keyCmd = mmd.getMap().getKeyClassMetaData(clr, op.getExecutionContext().getMetaDataManager());
                AbstractClassMetaData valCmd = mmd.getMap().getValueClassMetaData(clr, op.getExecutionContext().getMetaDataManager());

                StringBuffer cellValue = new StringBuffer("[");
                Map map = (Map)value;
                Iterator<Map.Entry> mapIter = map.entrySet().iterator();
                while (mapIter.hasNext())
                {
                    Map.Entry entry = mapIter.next();
                    cellValue.append("[");
                    if (keyCmd != null)
                    {
                        Object keyPC = op.getExecutionContext().persistObjectInternal(entry.getKey(), op, fieldNumber, -1);
                        Object keyID = op.getExecutionContext().getApiAdapter().getIdForObject(keyPC);
                        cellValue.append(keyID.toString());
                    }
                    else
                    {
                        cellValue.append(entry.getKey());
                    }
                    cellValue.append("],[");
                    if (valCmd != null)
                    {
                        Object valPC = op.getExecutionContext().persistObjectInternal(entry.getValue(), op, fieldNumber, -1);
                        Object valID = op.getExecutionContext().getApiAdapter().getIdForObject(valPC);
                        cellValue.append(valID.toString());
                    }
                    else
                    {
                        cellValue.append(entry.getValue());
                    }
                    cellValue.append("]");
                    if (mapIter.hasNext())
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                cell.setCellValue(createHelper.createRichTextString(cellValue.toString()));
            }
            else if (mmd.hasArray())
            {
                StringBuffer cellValue = new StringBuffer("[");
                for (int i=0;i<Array.getLength(value);i++)
                {
                    Object element = Array.get(value, i);
                    Object elementPC = op.getExecutionContext().persistObjectInternal(element, op, fieldNumber, -1);
                    Object elementID = op.getExecutionContext().getApiAdapter().getIdForObject(elementPC);
                    cellValue.append(elementID.toString());
                    if (i < (Array.getLength(value)-1))
                    {
                        cellValue.append(",");
                    }
                }
                cellValue.append("]");
                CreationHelper createHelper = row.getSheet().getWorkbook().getCreationHelper();
                cell.setCellValue(createHelper.createRichTextString(cellValue.toString()));
            }
        }
    }
}